/***************************************************************************

  CTable.c

  The Table, Field and Index class

  (c) 2000-2003 Benot Minisini <gambas@users.sourceforge.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __CTABLE_C

#include <ctype.h>

#include "main.h"

#include "CTable.h"


static void stolower(char *s)
{
  register char c;
  
  for(;;)
  {
    c = *s;
    if (!c)
      return;
    *s++ = tolower(c);
  }
}


/***************************************************************************

  Field

***************************************************************************/

static void free_one_field(DB_FIELD *field)
{
  GB.FreeString(&field->name);
  GB.StoreVariant(NULL, &field->def);
  field->type = GB_T_NULL;
  field->length = 0;
}

static void free_field(CTABLE* _object)
{
  GB.FreeString(&THIS->field);
  free_one_field(&THIS->field_info);
}


static void set_field(CTABLE *_object, char *field)
{
  free_field(THIS);

  GB.NewString(&THIS->field, field, 0);

  THIS->driver->Field.Info(THIS->db->handle, THIS->table, THIS->field, &THIS->field_info);
}


static void add_field(CTABLE *_object, DB_FIELD *field)
{
  DB_FIELD **fp;

  fp = &THIS->new_fields;

  for(;;)
  {
    if (!*fp)
    {
      *fp = field;
      field->next = NULL;
      return;
    }

    fp = &((*fp)->next);
  }
}


static int check_type(long type)
{
  if (type == GB_T_BOOLEAN
      || type == GB_T_INTEGER
      || type == GB_T_FLOAT
      || type == GB_T_DATE
      || type == GB_T_STRING)
    return FALSE;

  GB.Error("Bad type");
  return TRUE;
}

BEGIN_METHOD(CFIELD_find, GB_STRING name)

  char *field = GB.ToZeroString(ARG(name));

  if (THIS->create
      || !THIS->driver->Field.Exist(THIS->db->handle, THIS->table, field))
  {
    GB.Error("Unknown field: &1.&2", THIS->table, field);
    return;
  }

  set_field(THIS, field);

END_METHOD

#if 0
BEGIN_METHOD(CFIELD_delete, GB_STRING name)

  char *field = GB.ToZeroString(ARG(name));

  if (THIS->create
      || !THIS->driver->Field.Exist(THIS->handle, THIS->table, field))
  {
    GB.Error("Unknown field: &1.&2", THIS->table, field);
    return;
  }

  THIS->driver->Field.Delete(THIS->handle, THIS->table, field);

END_METHOD
#endif

BEGIN_METHOD(CFIELD_create, GB_STRING name; GB_INTEGER type; GB_INTEGER length; GB_VARIANT def)

  char *field = GB.ToZeroString(ARG(name));
  bool exist = FALSE;
  DB_FIELD *fp, new_field;

  if (DB_CheckName(field, "field"))
    return;

  if (THIS->create)
  {
    for (fp = THIS->new_fields; fp; fp = fp->next)
    {
      if (!strcasecmp(fp->name, field))
      {
        exist = TRUE;
        break;
      }
    }
  }
  else
    exist = THIS->driver->Field.Exist(THIS->db->handle, THIS->table, field);

  if (exist)
  {
    GB.Error("Field already exists: &1.&2", THIS->table, field);
    return;
  }

  new_field.next = NULL;

  new_field.type = VARG(type);
  if (check_type(new_field.type))
    return;

  new_field.length = VARGOPT(length, 0);
  if (new_field.length < 0)
    new_field.length = 0;
  else if (new_field.length > 65535)
    new_field.length = 65535;

  new_field.def._object.type = GB_T_NULL;

  if (!MISSING(def))
  {
    /*if (GB.Conv((GB_VALUE *)ARG(def), new_field.type))
      return;*/
    GB.StoreVariant(ARG(def), &new_field.def);
  }

  GB.NewString(&new_field.name, STRING(name), LENGTH(name));
  stolower(new_field.name);

  GB.Alloc((void **)&fp, sizeof(DB_FIELD));
  *fp = new_field;
  add_field(THIS, fp);

END_METHOD


BEGIN_METHOD_VOID(CFIELD_next)

  int *pos = (int *)GB.GetEnum();

  if (!THIS->create)
  {
    if (*pos == 0)
    {
      DB_FreeStringArray(&THIS->fields);
      THIS->driver->Field.List(THIS->db->handle, THIS->table, &THIS->fields);
    }

    if (THIS->fields)
    {
      if (*pos < GB.Count(THIS->fields))
      {
        set_field(THIS, THIS->fields[*pos]);
        (*pos)++;
        return;
      }
    }
  }

  GB.StopEnum();

END_METHOD


BEGIN_PROPERTY(CFIELD_name)

  GB.ReturnString(THIS->field);

END_PROPERTY


BEGIN_PROPERTY(CFIELD_type)

  GB.ReturnInteger(THIS->field_info.type);

END_PROPERTY


BEGIN_PROPERTY(CFIELD_length)

  GB.ReturnInteger(THIS->field_info.length);

END_PROPERTY


BEGIN_PROPERTY(CFIELD_default)

  GB.ReturnPtr(GB_T_VARIANT, &THIS->field_info.def);

END_PROPERTY


GB_DESC CTableFieldDesc[] =
{
  GB_DECLARE(".TableField", 0), GB_VIRTUAL_CLASS(),

  GB_METHOD("_call", NULL, CFIELD_find, "(Name)s"),
  GB_METHOD("Find", NULL, CFIELD_find, "(Name)s"),
  GB_METHOD("Create", NULL, CFIELD_create, "(Name)s(Type)i[(Length)i(Default)v]"),
  /*GB_METHOD("Delete", NULL, CFIELD_delete, "(Name)s"),*/

  GB_METHOD("_next", NULL, CFIELD_next, NULL),

  GB_PROPERTY_READ("Name", "s", CFIELD_name),
  GB_PROPERTY_READ("Type", "i", CFIELD_type),
  GB_PROPERTY_READ("Length", "i", CFIELD_length),
  GB_PROPERTY_READ("Default", "v", CFIELD_default),

  //GB_METHOD("Update", NULL, CTABLE_update, NULL),

  GB_END_DECLARE
};


/***************************************************************************

  Index

***************************************************************************/

static void free_index(CTABLE *_object)
{
  GB.FreeString(&THIS->index);
  GB.FreeString(&THIS->index_info.name);
  GB.FreeString(&THIS->index_info.fields);
  THIS->index_info.unique = FALSE;
}

static void set_index(CTABLE *_object, char *index)
{
  GB.FreeString(&THIS->index);
  free_index(THIS);

  GB.NewString(&THIS->index, index, 0);
  THIS->driver->Index.Info(THIS->db->handle, THIS->table, index, &THIS->index_info);
}


static bool check_index(CTABLE *_object, char *index, bool must_exist)
{
  bool exist = THIS->driver->Index.Exist(THIS->db->handle, THIS->table, index);

  if (must_exist)
  {
    if (!exist)
    {
      GB.Error("Unknown index: &1.&2", THIS->table, index);
      return TRUE;
    }
  }
  else
  {
    if (exist)
    {
      GB.Error("Index already exists: &1.&2", THIS->table, index);
      return TRUE;
    }
  }

  return FALSE;
}


BEGIN_METHOD(CINDEX_find, GB_STRING name)

  char *index = GB.ToZeroString(ARG(name));

  if (check_index(THIS, index, TRUE))
    return;

  set_index(THIS, index);

END_METHOD


BEGIN_METHOD(CINDEX_delete, GB_STRING name)

  char *index = GB.ToZeroString(ARG(name));

  if (check_index(THIS, index, TRUE))
    return;

  THIS->driver->Index.Delete(THIS->db->handle, THIS->table, index);

END_METHOD


BEGIN_METHOD(CINDEX_create, GB_STRING name; GB_STRING fields; GB_BOOLEAN unique)

  char *index = GB.ToZeroString(ARG(name));
  DB_INDEX info;

  if (DB_CheckName(index, "index"))
    return;

  if (check_index(THIS, index, FALSE))
    return;

  info.name = index;
  info.fields = GB.ToZeroString(ARG(fields));
  info.unique = VARG(unique);

  THIS->driver->Index.Create(THIS->db->handle, THIS->table, index, &info);

END_METHOD


BEGIN_METHOD_VOID(CINDEX_next)

  int *pos = (int *)GB.GetEnum();

  if (*pos == 0)
  {
    DB_FreeStringArray(&THIS->indexes);
    THIS->driver->Index.List(THIS->db->handle, THIS->table, &THIS->indexes);
  }

  if (THIS->indexes)
  {
    if (*pos < GB.Count(THIS->indexes))
    {
      set_index(THIS, THIS->indexes[*pos]);
      (*pos)++;
      return;
    }
  }

  GB.StopEnum();

END_METHOD


BEGIN_PROPERTY(CINDEX_name)

  GB.ReturnString(THIS->index);

END_PROPERTY


BEGIN_PROPERTY(CINDEX_fields)

  GB.ReturnString(THIS->index_info.fields);

END_PROPERTY


BEGIN_PROPERTY(CINDEX_unique)

  GB.ReturnBoolean(THIS->index_info.unique);

END_PROPERTY


BEGIN_PROPERTY(CINDEX_primary)

  GB.ReturnBoolean(THIS->index_info.primary);

END_PROPERTY


GB_DESC CTableIndexDesc[] =
{
  GB_DECLARE(".TableIndex", 0), GB_VIRTUAL_CLASS(),

  GB_METHOD("_call", NULL, CINDEX_find, "(Name)s"),
  GB_METHOD("Find", NULL, CINDEX_find, "(Name)s"),
  GB_METHOD("Create", NULL, CINDEX_create, "(Name)s(Fields)s[(Unique)b]"),
  GB_METHOD("Delete", NULL, CINDEX_delete, "(Name)s"),

  GB_METHOD("_next", NULL, CINDEX_next, NULL),

  GB_PROPERTY_READ("Name", "s", CINDEX_name),
  GB_PROPERTY_READ("Unique", "b", CINDEX_unique),
  GB_PROPERTY_READ("Primary", "b", CINDEX_primary),
  GB_PROPERTY_READ("Fields", "s", CINDEX_fields),

  //GB_METHOD("Update", NULL, CINDEX_update, NULL),

  GB_END_DECLARE
};


/***************************************************************************

  Table

***************************************************************************/

static int valid_table(CTABLE *_object)
{
  return (THIS->db->handle == NULL);
}

static void free_new_fields(CTABLE *_object)
{
  DB_FIELD *field;
  DB_FIELD *next;

  for (field = THIS->new_fields; field; field = next)
  {
    next = field->next;
    free_one_field(field);
    GB.Free((void **)&field);
  }

  THIS->new_fields = NULL;
}

static void free_table(CTABLE *_object)
{
  GB.FreeString(&THIS->table);
  DB_FreeStringArray(&THIS->primary);

  free_new_fields(THIS);

  free_field(THIS);
  free_index(THIS);

  DB_FreeStringArray(&THIS->fields);
  DB_FreeStringArray(&THIS->indexes);

  GB.Unref((void **)&THIS->db);
}


BEGIN_PROPERTY(CTABLE_primary_key)

  GB_ARRAY primary;
  int i, n;

  if (THIS->create)
  {
    if (READ_PROPERTY)
    {
      if (!THIS->primary)
      {
        GB.ReturnNull();
        return;
      }

      GB.ReturnObject(DB_StringArrayToGambasArray(THIS->primary));
    }
    else
    {
      DB_FreeStringArray(&THIS->primary);

      primary = (GB_ARRAY)VPROP(GB_OBJECT);
      n = GB.Array.Count(primary);

      GB.NewArray(&THIS->primary, sizeof(char *), n);

      for (i = 0; i < n; i++)
        GB.NewString(&THIS->primary[i], *((char **)GB.Array.Get(primary, i)), 0);
    }
  }
  else
  {
    if (READ_PROPERTY)
    {
      if (THIS->driver->Table.PrimaryKey(THIS->db->handle, THIS->table, &THIS->primary))
        return;

      GB.ReturnObject(DB_StringArrayToGambasArray(THIS->primary));
      DB_FreeStringArray(&THIS->primary);
    }
    else
      GB.Error("Read-only property");
  }

END_PROPERTY


BEGIN_PROPERTY(CTABLE_name)

  GB.ReturnString(THIS->table);

END_PROPERTY


BEGIN_PROPERTY(CTABLE_system)

  GB.ReturnBoolean(THIS->driver->Table.IsSystem(THIS->db->handle, THIS->table));

END_PROPERTY


BEGIN_METHOD_VOID(CTABLE_update)

  if (!THIS->new_fields)
  {
    GB.Error("No field");
    return;
  }

  if (!THIS->primary || GB.Count(THIS->primary) == 0)
  {
    GB.Error("No primary key");
    return;
  }

  if (THIS->driver->Table.Create(THIS->db->handle, THIS->table, THIS->new_fields, THIS->primary))
    return;

  free_new_fields(THIS);
  DB_FreeStringArray(&THIS->primary);
  THIS->create = FALSE;

END_METHOD


BEGIN_METHOD_VOID(CTABLE_free)

  free_table(THIS);

END_METHOD



GB_DESC CTableDesc[] =
{
  GB_DECLARE("Table", sizeof(CTABLE)), GB_NOT_CREATABLE(),

  GB_HOOK_CHECK(valid_table),

  GB_METHOD("_free", NULL, CTABLE_free, NULL),

  //GB_METHOD("AddField", NULL, CTABLE_add_field, "(Name)s(Type)i[(Length)i(Default)v"])

  GB_PROPERTY("PrimaryKey", "String[]", CTABLE_primary_key),

  GB_PROPERTY_READ("Name", "s", CTABLE_name),
  GB_PROPERTY_READ("System", "b", CTABLE_system),

  GB_METHOD("Update", NULL, CTABLE_update, NULL),

  GB_PROPERTY_SELF("Field", ".TableField"),
  GB_PROPERTY_SELF("Index", ".TableIndex"),

  GB_END_DECLARE
};
