/***************************************************************************

  CHttpClient.c

  Advanced Network component

  (c) 2003-2004 Daniel Campos Fernández <danielcampos@netcourrier.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/
#define __CHTTPCLIENT_C

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <curl/curl.h>
#include <curl/easy.h>
#include <curl/multi.h>

#include "main.h"
#include "gambas.h"
#include "CHttpClient.h"
#include "CProxy.h"



/****************************************************
 An array to manage all easy_curl pointers from
 each HTTPClient
*****************************************************/
void  **http_info;
int http_ninfo;
/*****************************************************
 CURLM : a pointer to use curl_multi interface,
 allowing asynchrnous work without using threads
 in this class. Here also a pipe will be stablished
 to link with Gambas watching interface
 ******************************************************/
CURLM *http_multicurl;
int http_pipe[2];
/******************************************************
 Events from this class
 ******************************************************/
DECLARE_EVENT (HTTP_FINISHED);
DECLARE_EVENT (HTTP_ERROR);
DECLARE_EVENT (HTTP_CONNECT);
DECLARE_EVENT (HTTP_READ);

GB_STREAM_DESC HttpStream = {
	CHttp_stream_open,
	CHttp_stream_close,
	CHttp_stream_read,
	CHttp_stream_write,
	CHttp_stream_seek,
	CHttp_stream_tell,
	CHttp_stream_flush,
	CHttp_stream_eof,
	CHttp_stream_lof
};
////////////////////////////////////////////////////////////////////
//	STREAM							  //
////////////////////////////////////////////////////////////////////

/* not allowed stream methods */
int CHttp_stream_open(GB_STREAM *stream, const char *path, int mode, void *data){return -1;}
int CHttp_stream_seek(GB_STREAM *stream, long pos, int whence){	return -1;}
int CHttp_stream_tell(GB_STREAM *stream, long *pos){return -1; }
int CHttp_stream_flush(GB_STREAM *stream) {	return 0;}
int CHttp_stream_close(GB_STREAM *stream) { return -1;}
int CHttp_stream_write(GB_STREAM *stream, char *buffer, long len){return -1;}

int CHttp_stream_lof(GB_STREAM *stream, long *len)
{
	CHTTPCLIENT *mythis;
	*len=0;

	mythis=(CHTTPCLIENT*)stream->_free[0];
	if ((mythis->iStatus !=4 ) && (mythis->iStatus != 0)) return -1;
	*len=mythis->len_data;
	return 0;
}
int CHttp_stream_eof(GB_STREAM *stream)
{
	CHTTPCLIENT *mythis;
	
	mythis=(CHTTPCLIENT*)stream->_free[0];
	if ((mythis->iStatus !=4 ) && (mythis->iStatus != 0)) return -1;
	if (!mythis->len_data) return -1;
	return 0;
}

int CHttp_stream_read(GB_STREAM *stream, char *buffer, long len)
{
	CHTTPCLIENT *mythis;
	
	mythis=(CHTTPCLIENT*)stream->_free[0];
	if ((mythis->iStatus !=4 ) && (mythis->iStatus != 0)) return -1;
	if (mythis->len_data < len) return -1;

	memcpy(buffer,mythis->buf_data,len);

	if (mythis->len_data == len)
	{
		mythis->len_data=0;
		GB.Free((void**)&mythis->buf_data);
		return 0;
	}

	mythis->len_data-=len;
	memmove(mythis->buf_data,len+mythis->buf_data,mythis->len_data);
	GB.Realloc((void**)&mythis->buf_data,mythis->len_data);

	return 0;
}


////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////

/*******************************************************************
####################################################################
	POSTED FUNCTIONS TO RAISE EVENTS
####################################################################
********************************************************************/
void http_raise_finished(long lParam)
{
	void *mythis;
	mythis=(void*)lParam;
	GB.Raise(mythis,HTTP_FINISHED,0);
	GB.Unref(&mythis);
}
void http_raise_error(long lParam)
{
	void *mythis;
	mythis=(void*)lParam;
	GB.Raise(mythis,HTTP_ERROR,0);
	GB.Unref(&mythis);
}
void http_raise_connect(long lParam)
{
	void *mythis;
	mythis=(void*)lParam;
	GB.Raise(mythis,HTTP_CONNECT,0);
	GB.Unref(&mythis);
}
void http_raise_read(long lParam)
{
	void *mythis;
	mythis=(void*)lParam;
	GB.Raise(mythis,HTTP_READ,0);
	GB.Unref(&mythis);
}

/*******************************************************************
####################################################################
	FUNCTIONS TO MANAGE http_info array of structures
####################################################################
********************************************************************/
int http_find_info (CURL *curlfind)
{
	CHTTPCLIENT *mythis;
	int pos=-1;
	int myloop;

	if ( !http_ninfo ) return -1;
	if (!curlfind) return -1;

	for (myloop=0; (myloop<http_ninfo) && (pos==-1) ;myloop++)
	{
		mythis=http_info[myloop];
		if ( (mythis->curlhandle == curlfind) && (mythis->curlhandle) ) pos=myloop;
	}
	return pos;

}

/*******************************************************************
####################################################################
	CALLBACKS FROM CURL LIBRARY
####################################################################
********************************************************************/
void http_parse_header(CHTTPCLIENT *mythis)
{
	/* Obtains HTTP return code and return string */
	int myloop;
	int mypos=0;
	int npos=0;
	int nposend=0;
	int len;
	char *buf;

	len=strlen(mythis->buf_header[0]);
    buf=mythis->buf_header[0];

	for (myloop=4;myloop<len;myloop++)
	{
		if (buf[myloop]==' ')
		{
			mypos=myloop+1;
			break;
		}
	}
	if (!mypos) return;

	for (myloop=mypos;myloop<len;myloop++)
	{
		if (buf[myloop]==' ')
		{
			if (npos)
			{
				npos=myloop+1;
				break;
			}
		}
		else
		{
			if ( !((buf[myloop]>47) && (buf[myloop]<58)) )
			{
				if (buf[myloop] != 13)
					return;
				else
					return;
			}
			else
			{
				npos++;
				if (npos>3) return;
				mythis->ReturnCode*=10;
				mythis->ReturnCode+=(buf[myloop]-48);

			}
		}
	}

	GB.Alloc((void**)&mythis->ReturnString,(nposend+1)*sizeof(char));
	mythis->ReturnString[nposend]=0;
	for (myloop=npos;myloop<(nposend+npos);myloop++)
		mythis->ReturnString[myloop-npos]=buf[myloop];

}

int http_header_curl(void *buffer, size_t size, size_t nmemb, void *c_handle)
{
	/* this call-back is used to receive each line of header */
	int pos;
	CHTTPCLIENT *mythis;

	pos=http_find_info((CURL*)c_handle);
	if (pos<0) return 0;

	mythis=(CHTTPCLIENT*)http_info[pos];

	if (!mythis->len_header)
	{
		GB.Alloc((void**)&mythis->buf_header,sizeof(char*));
		GB.Alloc((void**)&mythis->buf_header[0],nmemb+1);
	}
	else
	{
		GB.Realloc((void**)&mythis->buf_header,sizeof(char*)*(1+mythis->len_header));
		GB.Alloc((void**)&(mythis->buf_header[mythis->len_header]),nmemb+1);
		mythis->buf_header[mythis->len_header][nmemb]=0;
	}
	strncpy(mythis->buf_header[mythis->len_header],buffer,nmemb);
	mythis->len_header++;

	if (mythis->iStatus==6)
	{
		mythis->iStatus=4;
		GB.Ref(mythis);
		GB.Post(http_raise_connect,(long)mythis);
	}

	return nmemb;
}



int http_write_curl(void *buffer, size_t size, size_t nmemb, void *c_handle)
{
	/* this callback is used to receive each piece of body */
	int pos;
	CHTTPCLIENT *mythis;

	pos=http_find_info((CURL*)c_handle);

	if (pos<0) return 0;

	mythis=(CHTTPCLIENT*)http_info[pos];
	mythis->iStatus=4;

	if (!mythis->ReturnCode) http_parse_header(mythis);

	if (mythis->c_file)
	{
		return fwrite(buffer,size,nmemb,mythis->c_file);
	}
	else
	{
		if (!mythis->len_data)
			GB.Alloc((void**)&mythis->buf_data,nmemb);
		else
			GB.Realloc((void**)&mythis->buf_data,nmemb+mythis->len_data);
		memcpy(mythis->buf_data+mythis->len_data,buffer,nmemb);
		mythis->len_data+=nmemb;
	}

	GB.Ref(mythis);
	GB.Post(http_raise_read,(long)mythis);
	return nmemb;
}

void http_manage_errcode(CHTTPCLIENT *mythis,long ErrCode)
{
	if (mythis->c_file) fclose(mythis->c_file);
	switch ( ErrCode )
	{
		case CURLE_OK:
			curl_multi_remove_handle(http_multicurl,mythis->curlhandle);
			mythis->iStatus=0;
			GB.Ref(mythis);
			GB.Post(http_raise_finished,(long)mythis);
			break;
        	default:
			curl_multi_remove_handle(http_multicurl,mythis->curlhandle);
			Adv_remove_info(&http_info, &http_ninfo, (void*)mythis);

			mythis->iStatus=-1*(1000+ErrCode);
			GB.Ref(mythis);
			GB.Post(http_raise_error,(long)mythis);
			break;
	}

}
/***************************************************************
 This CallBack is called each event loop by Gambas to test
 the status of curl descriptors
 ***************************************************************/
void http_post_curl(long lParam)
{
	CURLMsg *Msg;
	int nread;
	int pos;
	char buf[1];
	CHTTPCLIENT *mythis;

	read (http_pipe[0],buf,sizeof(char));
	while(CURLM_CALL_MULTI_PERFORM == curl_multi_perform(http_multicurl,&nread));

	if (!nread)
	{
		GB.Watch(http_pipe[0],GB_WATCH_NONE,http_post_curl,0);
		close(http_pipe[0]);
		close(http_pipe[1]);
		http_pipe[0]=-1;
		http_pipe[1]=-1;
	}

	do
	{
		Msg=curl_multi_info_read(http_multicurl,&nread);
		if (!Msg) nread=0;
		if (Msg)
		{
			pos=http_find_info (Msg->easy_handle);
			if ( pos > -1 )
			{
				if ( (mythis=(CHTTPCLIENT*)http_info[pos])==NULL) return;
				http_manage_errcode(mythis,Msg->data.result);
			}
		}
	} while (nread);

	write(http_pipe[1],"1",sizeof(char));
}
////////////////////////////////////////////////////////////////////////
//###################### PROXY related properties ######################
////////////////////////////////////////////////////////////////////////
/*********************************************
 Proxy name and port ( Host:Port format)
 *********************************************/
BEGIN_PROPERTY (CHttpClient_Proxy)

	GB.ReturnObject(THIS->proxy);

END_PROPERTY



////////////////////////////////////////////////////////////////////////
//########################## Other properties ##########################
////////////////////////////////////////////////////////////////////////
/*****************************************************************
 URL to work with
 *****************************************************************/


BEGIN_PROPERTY ( CHttpClient_URL )

	if (READ_PROPERTY)
	{
		GB.ReturnNewString(THIS->sURL,0);
		return;
	}

	if (THIS->iStatus > 0)
	{
		GB.Error ("URL property can not be changed while working");
		return;
  	}

	if (THIS->sURL) GB.Free((void**)&THIS->sURL);
	GB.Alloc((void**)&THIS->sURL,(strlen(GB.ToZeroString(PROP(GB_STRING)))+1)*sizeof(char));
	strcpy(THIS->sURL,GB.ToZeroString(PROP(GB_STRING)));
	Adv_correct_url(&THIS->sURL,"http://");

END_PROPERTY

////////////////////////////////////////////////////////////////////////
//########################## Cookies ###################################
////////////////////////////////////////////////////////////////////////
BEGIN_PROPERTY ( CHttpClient_UpdateCookies )

	if (READ_PROPERTY)
	{
		GB.ReturnBoolean(THIS->updatecookies);
		return;
	}

	if (THIS->iStatus > 0)
	{
		GB.Error ("UpdateCookies property can not be changed while working");
		return;
  	}

	if (VPROP(GB_BOOLEAN))
		THIS->updatecookies=1;
	else
		THIS->updatecookies=0;

END_PROPERTY

BEGIN_PROPERTY ( CHttpClient_CookiesFile )

	if (READ_PROPERTY)
	{
		GB.ReturnNewString(THIS->cookiesfile,0);
		return;
	}

	if (THIS->iStatus > 0)
	{
		GB.Error ("CookiesFile property can not be changed while working");
		return;
  	}

	if (THIS->cookiesfile)
	{
		GB.Free((void**)&THIS->cookiesfile);
		THIS->cookiesfile=NULL;
	}
	if ( strlen(GB.ToZeroString(PROP(GB_STRING))) )
	{
		GB.Alloc( (void**)&THIS->cookiesfile, \
		          (strlen(GB.ToZeroString(PROP(GB_STRING)))+1)*sizeof(char));
		strcpy(THIS->cookiesfile,GB.ToZeroString(PROP(GB_STRING)));
	}

END_PROPERTY
/*********************************************
 HTTP User ( User:Password format )
 *********************************************/
BEGIN_PROPERTY ( CHttpClient_sUser )

	if (READ_PROPERTY)
	{
		GB.ReturnString(THIS->user.user);
		return;
	}

	if (THIS->iStatus > 0)
	{
		GB.Error ("User property can not be changed while working");
		return;
  	}

	if ( THIS->user.user ) GB.FreeString ( &(THIS->user.user) );
	GB.StoreString(PROP(GB_STRING), &(THIS->user.user) );


END_PROPERTY
BEGIN_PROPERTY ( CHttpClient_Password )


	if (READ_PROPERTY)
	{
		GB.ReturnString(THIS->user.pwd);
		return;
	}

	if (THIS->iStatus > 0)
	{
		GB.Error ("User property can not be changed while working");
		return;
  	}

	if ( THIS->user.pwd ) GB.FreeString ( &(THIS->user.pwd) );
	GB.StoreString(PROP(GB_STRING), &(THIS->user.pwd) );


END_PROPERTY
/***********************************************
 HTTP authentication : Plain, GSS negotiation, Digest or NTLM
************************************************/
BEGIN_PROPERTY (CHttpClient_Auth)

	if (READ_PROPERTY)
	{
		GB.ReturnInteger(THIS->auth);
		return;
	}

	if (THIS->iStatus > 0)
	{
		GB.Error ("Auth property can not be changed while working");
		return;
  	}

	if (Adv_user_SETAUTH (&THIS->user,VPROP(GB_INTEGER)) )
		GB.Error ("Unknown authentication method");


END_PROPERTY
/*********************************************
 User Agent string to be sent to the server
 *********************************************/
BEGIN_PROPERTY ( CHttpClient_UserAgent )

	if (READ_PROPERTY)
	{
		GB.ReturnString(THIS->sUserAgent);
		return;
	}

	if (THIS->iStatus > 0)
	{
		GB.Error ("UserAgent property can not be changed while working");
		return;
  	}
	GB.StoreString(PROP(GB_STRING), &THIS->sUserAgent);


END_PROPERTY
/*********************************************
 Status : inactive, working or Error code
 *********************************************/
BEGIN_PROPERTY ( CHttpClient_Status )

	GB.ReturnInteger(THIS->iStatus);

END_PROPERTY
/*********************************************
 Return code received from Http Server
 *********************************************/
BEGIN_PROPERTY ( CHttpClient_ReturnCode )

	GB.ReturnInteger(THIS->ReturnCode);

END_PROPERTY
/*********************************************
 Return string received from Http Server
 *********************************************/
BEGIN_PROPERTY ( CHttpClient_ReturnString )

	GB.ReturnNewString(THIS->ReturnString,0);

END_PROPERTY

/*********************************************
 Header of data received from Http Server
 *********************************************/
BEGIN_PROPERTY ( CHttpClient_Headers )

	GB_ARRAY retval;
	char *element;
	int myloop;

	if ( (THIS->iStatus==4) || (THIS->iStatus==0) )
	{
		if (THIS->len_header)
		{
			GB.Array.New(&retval,GB_T_STRING,THIS->len_header);
			for (myloop=0;myloop<THIS->len_header;myloop++)
			{
			  GB.NewString(&element,THIS->buf_header[myloop],strlen(THIS->buf_header[myloop]));
			  *((char **)GB.Array.Get(retval,myloop)) = element;
			}
			GB.ReturnObject ( retval );
			return;
		}

	}


END_PROPERTY

//*************************************************************************
//#################### INITIALIZATION AND DESTRUCTION #####################
//*************************************************************************
/*************************************************
 Gambas object "Constructor"
 *************************************************/
BEGIN_METHOD_VOID(CHTTPCLIENT_new)

	
	THIS->stream.desc=NULL;
	THIS->curlhandle=NULL;
	THIS->iStatus=0;

	THIS->sUserAgent=NULL;
	THIS->sURL=NULL;
	GB.Alloc((void**)&THIS->sURL,sizeof(char)*(1+strlen("http://127.0.0.1:80")));
	strcpy(THIS->sURL,"http://127.0.0.1:80");
	GB.NewString(&THIS->sUserAgent,"Gambas Http/1.0",0);

	THIS->buf_header=NULL;
	THIS->buf_data=NULL;
	THIS->ReturnCode=0;
	THIS->ReturnString=NULL;

	THIS->cookiesfile=NULL;

	THIS->sContentType=NULL;
	THIS->sPostData=NULL;

	THIS->c_file=NULL;

	Adv_user_NEW  (&THIS->user);
	
	GB.New ((void**)&THIS->proxy,GB.FindClass(".Proxy"),NULL,NULL);
	GB.Ref((void*)THIS->proxy);
	Adv_proxy_NEW(&THIS->proxy->proxy);
	THIS->proxy->parent_status=&THIS->iStatus;
	return;


END_METHOD

/*************************************************
 Gambas object "Destructor"
 *************************************************/
void http_free_all(CHTTPCLIENT *mythis)
{
	if (mythis->c_file) fclose(mythis->c_file);
	http_reset(mythis);
	if (mythis->curlhandle) curl_easy_cleanup(mythis->curlhandle);
	if (mythis->sUserAgent) GB.FreeString(&mythis->sUserAgent);
	if (mythis->sURL)	GB.Free((void**)&mythis->sURL);
	if (mythis->cookiesfile)	GB.Free((void**)&mythis->cookiesfile);
	if (mythis->ReturnString) GB.Free((void**)&mythis->ReturnString);
	Adv_user_CLEAR  (&mythis->user);
	Adv_proxy_CLEAR(&mythis->proxy->proxy);
	GB.Unref((void**)&mythis->proxy);
	//GB.Free((void**)&mythis->proxy);
}

BEGIN_METHOD_VOID(CHTTPCLIENT_free)

	http_free_all(THIS);

END_METHOD

BEGIN_METHOD_VOID(CHTTPCLIENT_init)

	http_info=NULL;
	http_ninfo=0;
	http_pipe[0]=-1;
	http_pipe[1]=-1;
	http_multicurl=curl_multi_init();

END_METHOD

BEGIN_METHOD_VOID(CHTTPCLIENT_exit)

	curl_multi_cleanup(http_multicurl);

END_METHOD

//*************************************************************************
//#################### METHODS ############################################
//*************************************************************************
/*************************************************
 GET HTTP method
 *************************************************/
void initialize_curl_handle(CHTTPCLIENT *mythis,char *url)
{
	if (mythis->curlhandle)
	{
		if (Adv_Comp ( mythis->user.userpwd,mythis->user.user,mythis->user.pwd))
			http_stop(mythis);
	}

	if (!mythis->curlhandle) mythis->curlhandle=curl_easy_init();

	//curl_easy_setopt(mythis->curlhandle, CURLOPT_VERBOSE,1);
	curl_easy_setopt(mythis->curlhandle, CURLOPT_USERAGENT,mythis->sUserAgent);
	curl_easy_setopt(mythis->curlhandle, CURLOPT_HEADERFUNCTION, http_header_curl);
	curl_easy_setopt(mythis->curlhandle, CURLOPT_WRITEFUNCTION, http_write_curl);
	curl_easy_setopt(mythis->curlhandle, CURLOPT_FILE, mythis->curlhandle);
	curl_easy_setopt(mythis->curlhandle, CURLOPT_WRITEHEADER, mythis->curlhandle);
	curl_easy_setopt(mythis->curlhandle, CURLOPT_COOKIEFILE, mythis->cookiesfile);
	if (mythis->updatecookies)
		curl_easy_setopt(mythis->curlhandle, CURLOPT_COOKIEJAR, mythis->cookiesfile);
	else
		curl_easy_setopt(mythis->curlhandle, CURLOPT_COOKIEJAR, NULL);

	Adv_proxy_SET (&mythis->proxy->proxy,mythis->curlhandle);
	Adv_user_SET  (&mythis->user, mythis->curlhandle);
	curl_easy_setopt(mythis->curlhandle, CURLOPT_URL,url);

	mythis->ReturnCode=0;
	if (mythis->ReturnString)
	{
		GB.Free((void**)&mythis->ReturnString);
		mythis->ReturnString=NULL;
	}
	http_reset(mythis);
	mythis->iStatus=6;
	mythis->stream.desc=&HttpStream;
	mythis->stream._free[0]=(long)mythis;
}


int http_get (CHTTPCLIENT *mythis,char *url)
{
	if (mythis->iStatus > 0) return 1;

	mythis->iMethod=0;
	initialize_curl_handle(mythis,url);
	curl_easy_setopt(mythis->curlhandle,CURLOPT_HTTPGET,1);
	curl_multi_add_handle(http_multicurl,mythis->curlhandle);
	//http_add_info(mythis);
	Adv_add_info(&http_info, &http_ninfo, (void*)mythis);
	if (http_pipe[0]==-1)
	{
		pipe(http_pipe);
		GB.Watch(http_pipe[0],GB_WATCH_READ,http_post_curl,0);
		write(http_pipe[1],"1",sizeof(char));
	}
	return 0;
}

BEGIN_METHOD(CHTTPCLIENT_Get,GB_STRING TargetHost;)

	if (!MISSING(TargetHost))
	{
		if (THIS->iStatus > 0)
		{
			GB.Error("Still active");
			return;
		}
		THIS->c_file=fopen(STRING(TargetHost),"w");
		if (!THIS->c_file)
		{
			GB.Error("Unable to open file for writing");
			return;
		}
	}

	if (http_get(THIS,THIS->sURL))
	{
		GB.Error("Still active");
		return;
	}

END_METHOD

/*************************************************
 POST HTTP method
 *************************************************/
int http_post (CHTTPCLIENT *mythis,char *sContent,char *sData,int lendata,char *url)
{
	int mylen;
	struct curl_slist *headers=NULL;
	int myloop;

	if (mythis->iStatus > 0) return 1;
	if (!sContent) return 2;
	if (!sData) return 3;
	for (myloop=0;myloop<strlen(sContent);myloop++)
		if (sContent[myloop]<32) return 1;


	initialize_curl_handle(mythis,url);

    	mylen=strlen(sContent) + strlen("Content-Type: ") + 1;
	GB.Alloc((void*)&mythis->sContentType,mylen);
	GB.Alloc((void*)&mythis->sPostData,lendata+1);
	strncpy(mythis->sPostData,sData,lendata);
	mythis->sContentType[0]=0;
	strcpy(mythis->sContentType,"Content-Type: ");
	strcat(mythis->sContentType,sContent);

	mythis->iMethod=1;
	headers=curl_slist_append(headers,mythis->sContentType );
	curl_easy_setopt(mythis->curlhandle,CURLOPT_POSTFIELDS,mythis->sPostData);
	curl_easy_setopt(mythis->curlhandle,CURLOPT_POSTFIELDSIZE,lendata);
	curl_easy_setopt(mythis->curlhandle,CURLOPT_HTTPHEADER,headers);

	curl_multi_add_handle(http_multicurl,mythis->curlhandle);
	//http_add_info(mythis);
	Adv_add_info(&http_info, &http_ninfo, (void*)mythis);
	if (http_pipe[0]==-1)
	{
		pipe(http_pipe);
		GB.Watch(http_pipe[0],GB_WATCH_READ,http_post_curl,0);
		write(http_pipe[1],"1",sizeof(char));
	}
	return 0;
}

BEGIN_METHOD(CHTTPCLIENT_Post,GB_STRING sContentType;GB_STRING sData;GB_STRING TargetHost;)

	if (!MISSING(TargetHost))
	{
		if (THIS->iStatus > 0)
		{
			GB.Error("Still active");
			return;
		}
		THIS->c_file=fopen(STRING(TargetHost),"w");
		if (!THIS->c_file)
		{
			GB.Error("Unable to open file for writing");
			return;
		}
	}

	switch(http_post (THIS,STRING(sContentType),STRING(sData),LENGTH(sData),THIS->sURL) )
	{
		case 1:
			GB.Error("Still active");
			return;
		case 2:
			GB.Error("Invalid content type");
			return;
		case 3:
			GB.Error("Invalid data");
			return;
	}

END_METHOD


BEGIN_METHOD_VOID(CHTTPCLIENT_Peek)

	if ( (THIS->len_data) && (THIS->buf_data) )
	{
		GB.ReturnNewString(THIS->buf_data,THIS->len_data);
		return;
	}
	GB.ReturnNewString(NULL,0);

END_METHOD

void http_reset(CHTTPCLIENT *mythis)
{
	int myloop;

	if (mythis->buf_data)
	{
		GB.Free((void**)&mythis->buf_data);
		mythis->buf_data=NULL;
	}
	if (mythis->buf_header)
	{
		for (myloop=0;myloop < mythis->len_header;myloop++)
			GB.Free((void**)&mythis->buf_header[myloop]);
		GB.Free((void**)&mythis->buf_header);
		mythis->buf_header=NULL;
	}
	if (mythis->sContentType)
	{
		GB.Free((void**)&mythis->sContentType);
		mythis->sContentType=NULL;
	}
	if (mythis->sPostData)
	{
		GB.Free((void**)&mythis->sPostData);
		mythis->sPostData =NULL;
	}
	mythis->len_data=0;
	mythis->len_header=0;
}

void http_stop(CHTTPCLIENT *mythis)
{
	if (mythis->c_file) fclose(mythis->c_file);
	if (mythis->curlhandle)
	{
		curl_multi_remove_handle(http_multicurl,mythis->curlhandle);
		curl_easy_cleanup(mythis->curlhandle);
		//http_remove_info(mythis);
		Adv_remove_info(&http_info, &http_ninfo, (void*)mythis);
		mythis->curlhandle=NULL;
	}
	mythis->iStatus=0;
	http_reset(mythis);
}

BEGIN_METHOD_VOID(CHTTPCLIENT_Stop)

	http_stop(THIS);

END_METHOD

//*************************************************************************
//#################### GAMBAS INTERFACE ###################################
//*************************************************************************
GB_DESC CHttpClientDesc[] =
{

  GB_DECLARE("HttpClient", sizeof(CHTTPCLIENT)),

  GB_INHERITS(".Stream"),


  GB_STATIC_METHOD("_init",NULL,CHTTPCLIENT_init, NULL),
  GB_STATIC_METHOD("_exit",NULL,CHTTPCLIENT_exit, NULL),
  GB_METHOD("_new", NULL, CHTTPCLIENT_new, NULL),
  GB_METHOD("_free", NULL, CHTTPCLIENT_free, NULL),

  GB_METHOD("Stop", NULL, CHTTPCLIENT_Stop, NULL),
  GB_METHOD("Get", NULL, CHTTPCLIENT_Get, "[(TargetFile)s]"),
  GB_METHOD("Post", NULL, CHTTPCLIENT_Post, "(ContentType)s(Data)s[(TargetFile)s]"),
  GB_METHOD("Peek","s", CHTTPCLIENT_Peek, NULL),

  GB_EVENT("Finished", NULL, NULL, &HTTP_FINISHED),
  GB_EVENT("Connect", NULL, NULL, &HTTP_CONNECT),
  GB_EVENT("Read", NULL, NULL, &HTTP_READ),
  GB_EVENT("Error", NULL,NULL, &HTTP_ERROR),

  GB_PROPERTY("URL", "s",CHttpClient_URL),

  GB_PROPERTY("User","s",CHttpClient_sUser),
  GB_PROPERTY("Password","s",CHttpClient_Password),
  GB_PROPERTY("Auth","i<Net,AuthNone,AuthBasic,AuthNTLM,AuthDIGEST,AuthGSSNEGOTIATE>",CHttpClient_Auth),
  
  

  GB_PROPERTY_READ("Proxy",".Proxy",CHttpClient_Proxy),

  GB_PROPERTY("CookiesFile", "s",CHttpClient_CookiesFile),
  GB_PROPERTY("UpdateCookies", "b",CHttpClient_UpdateCookies),

  GB_PROPERTY("Headers", "String[]", CHttpClient_Headers),
  GB_PROPERTY("UserAgent", "s", CHttpClient_UserAgent),

  GB_PROPERTY_READ("Status","i",CHttpClient_Status),
  GB_PROPERTY_READ("Code","i",CHttpClient_ReturnCode),
  GB_PROPERTY_READ("Reason","s",CHttpClient_ReturnString),
  
  GB_CONSTANT("_Properties", "s", HTTP_PROPERTIES),
  GB_CONSTANT("_DefaultEvent", "s", "Read"),

  GB_END_DECLARE
};

