/***************************************************************************

  CContainer.cpp

  The Container class

  (c) 2000-2003 Benot Minisini <gambas@users.sourceforge.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __CCONTAINER_CPP



#include <qnamespace.h>
#include <qapplication.h>
#include <qmainwindow.h>
#include <qlayout.h>

#include "gambas.h"

#include "CWidget.h"
#include "CWindow.h"
#include "CConst.h"

#include "CContainer.h"

/*#define DEBUG*/

static QWidget *get_widget(QObjectList *list)
{
  for(;;)
  {
    QObject *ob = list->current();
    if (!ob)
      return NULL;
    list->next();
    if (ob->isWidgetType())
    {
      if (((QWidget *)ob)->isVisible())
        return (QWidget *)ob;
    }
  }
}

void CCONTAINER_arrange(QFrame *cont)
{
  static bool lock = false;
  QWidget *wid;
  int x, y, w, h, arr;
  int xc, yc, wc, hc;
  bool swap;

  if (lock)
    return;

  arr = ((CCONTAINER_ARRANGEMENT *)CWidget::get(cont))->arrangement;

  if (arr == ARRANGE_NONE)
    return;

  QObjectList *list = (QObjectList *)cont->children();

  if (!list || list->count() == 0)
    return;

  lock = true;

  list->first();

  swap = (arr & 1) == 0;

  xc = cont->contentsRect().x();
  yc = cont->contentsRect().y();
  wc = cont->contentsRect().width();
  hc = cont->contentsRect().height();

  switch (arr)
  {
    case ARRANGE_HORIZONTAL:
    case ARRANGE_VERTICAL:

      x = xc;
      y = yc;

      for(;;)
      {
        wid = get_widget(list);
        if (!wid)
          break;

        if (swap)
        {
          wid->setGeometry(x, y, wc, wid->height());
          y += wid->height();
        }
        else
        {
          wid->setGeometry(x, y, wid->width(), hc);
          x += wid->width();
        }
      }
      break;

    case ARRANGE_LEFT_RIGHT:
    case ARRANGE_TOP_BOTTOM:

      x = xc;
      y = yc;
      w = h = 0;

      for(;;)
      {
        wid = get_widget(list);
        if (!wid)
          break;

        if (swap)
        {
          if ((y > yc) && ((y + wid->height()) >= hc))
          {
            y = yc;
            x += w;
            w = 0;
          }

          wid->move(x, y);
          y += wid->height();

          if (wid->width() > w)
            w = wid->width();
        }
        else
        {
          if ((x > xc) && ((x + wid->width()) >= wc))
          {
            x = xc;
            y += h;
            h = 0;
          }

          wid->move(x, y);
          x += wid->width();

          if (wid->height() > h)
            h = wid->height();
        }
      }
      break;
  }

  lock = false;
}



/***************************************************************************

  class MyContainer

***************************************************************************/

MyContainer::MyContainer(QWidget *parent)
: QFrame(parent)
{
}


void MyContainer::frameChanged(void)
{
  QFrame::frameChanged();
  CCONTAINER_arrange(this);
}

void MyContainer::resizeEvent(QResizeEvent *e)
{
  QFrame::resizeEvent(e);
  CCONTAINER_arrange(this);
}

void MyContainer::childEvent(QChildEvent *e)
{
  QFrame::childEvent(e);

  if (!e->child()->isWidgetType())
    return;

  if (e->inserted())
  {
    e->child()->installEventFilter(this);
  }
  else if (e->removed())
  {
    e->child()->removeEventFilter(this);
  }

  CCONTAINER_arrange(this);
}

bool MyContainer::eventFilter(QObject *o, QEvent *e)
{
  int type = e->type();

  if (type == QEvent::Move || type == QEvent::Resize || type == QEvent::Show || type == QEvent::Hide)
    CCONTAINER_arrange(this);

  return QObject::eventFilter(o, e);
}


/***************************************************************************

  CContainer

***************************************************************************/

BEGIN_METHOD_VOID(CCONTAINER_next)

  #ifdef DEBUG
  if (!THIS_CONTAINER)
    qDebug("Null container");
  #endif

  QObjectList *list = (QObjectList *)THIS_CONTAINER->children();
  unsigned int index;
  CWIDGET *widget;

  for(;;)
  {
    index = ENUM(int);

    if (list == NULL || index >= list->count())
    {
      GB.StopEnum();
      return;
    }

    ENUM(int) = index + 1;

    widget = CWidget::getReal(list->at(index));
    if (widget)
    {
      GB.ReturnObject(widget);
      return;
    }
  }

END_METHOD


BEGIN_PROPERTY(CCONTAINER_count)

  #ifdef DEBUG
  if (!THIS_CONTAINER)
    qDebug("Null container");
  #endif

  QWidget *wid = THIS_CONTAINER;

  if (wid->children() == NULL)
    GB.ReturnInteger(0);
  else
    GB.ReturnInteger(wid->children()->count());

END_PROPERTY


BEGIN_PROPERTY(CCONTAINER_x)

  #ifdef DEBUG
  if (!THIS_CONTAINER)
    qDebug("Null container");
  #endif

  QPoint p(0, 0);
  /*if (WIDGET->layout())
    WIDGET->layout()->activate();*/
  GB.ReturnInteger(THIS_CONTAINER->mapTo(WIDGET, p).x());

END_PROPERTY


BEGIN_PROPERTY(CCONTAINER_y)

  #ifdef DEBUG
  if (!THIS_CONTAINER)
    qDebug("Null container");
  #endif

  QPoint p(0, 0);

  /*if (WIDGET->layout())
    WIDGET->layout()->activate();*/
  GB.ReturnInteger(THIS_CONTAINER->mapTo(WIDGET, p).y());

END_PROPERTY


BEGIN_PROPERTY(CCONTAINER_width)

  #ifdef DEBUG
  if (!THIS_CONTAINER)
    qDebug("Null container");
  #endif

  /*if (WIDGET->isTopLevel() && !WIDGET->isVisible())
    GB.ReturnInteger(((CWINDOW *)THIS)->w);
  else*/
  /*if (WIDGET->layout())
    WIDGET->layout()->activate();*/
  GB.ReturnInteger(THIS_CONTAINER->width());

END_PROPERTY


BEGIN_PROPERTY(CCONTAINER_height)

  #ifdef DEBUG
  if (!THIS_CONTAINER)
    qDebug("Null container");
  #endif

  /*if (WIDGET->isTopLevel() && !WIDGET->isVisible())
    GB.ReturnInteger(((CWINDOW *)THIS)->h);
  else*/
  /*if (WIDGET->layout())
    WIDGET->layout()->activate();*/
  GB.ReturnInteger(THIS_CONTAINER->height());

END_PROPERTY


BEGIN_PROPERTY(CCONTAINER_arrangement)

  int arr;

  if (READ_PROPERTY)
    GB.ReturnInteger(THIS_ARRANGEMENT);
  else
  {
    arr = VPROP(GB_INTEGER);
    if (arr < 0 || arr > 8)
      return;
    THIS_ARRANGEMENT = arr;
    CCONTAINER_arrange((QFrame *)THIS_CONTAINER);
  }

END_PROPERTY


GB_DESC CChildrenDesc[] =
{
  GB_DECLARE(".ContainerChildren", sizeof(CCONTAINER)), GB_VIRTUAL_CLASS(),

  GB_METHOD("_next", "Control", CCONTAINER_next, NULL),
  GB_PROPERTY_READ("Count", "i", CCONTAINER_count),

  GB_END_DECLARE
};


GB_DESC CContainerDesc[] =
{
  GB_DECLARE("Container", sizeof(CCONTAINER)), GB_INHERITS("Control"),

  GB_NOT_CREATABLE(),

  GB_PROPERTY_SELF("Children", ".ContainerChildren"),

  GB_PROPERTY_READ("ClientX", "i", CCONTAINER_x),
  GB_PROPERTY_READ("ClientY", "i", CCONTAINER_y),
  GB_PROPERTY_READ("ClientW", "i", CCONTAINER_width),
  GB_PROPERTY_READ("ClientWidth", "i", CCONTAINER_width),
  GB_PROPERTY_READ("ClientH", "i", CCONTAINER_height),
  GB_PROPERTY_READ("ClientHeight", "i", CCONTAINER_height),

  GB_END_DECLARE
};



