/***************************************************************************

  alloc_t.h

  Memory management routines template

  (c) 2000-2004 Benot Minisini <gambas@users.sourceforge.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __GB_ALLOC_C

#include <config.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <stdlib.h>

#include "gb_common.h"
#include "gb_error.h"
#include "gb_alloc.h"

#if DEBUG_MEMORY
# ifdef __GNU_LIBRARY__
#  include <mcheck.h>
# endif
#endif


#if MEMORY_OPTIM

#define MAX_KEEP 32
#define GRANULARITY 8

PRIVATE void *_keep[MAX_KEEP] = { 0 };
PRIVATE long _keep_malloc = 0;
PRIVATE long _keep_free = 0;


PRIVATE size_t round_size(size_t size)
{
  if (size & (GRANULARITY - 1))
    size = (size & ~(GRANULARITY - 1)) + GRANULARITY;

  return size;
}

#endif

PUBLIC long MEMORY_count = 0;

PUBLIC void MEMORY_init(void)
{
  /*mcheck(NULL);*/
#if DEBUG_MEMORY
# ifdef __GNU_LIBRARY__
  mtrace();
  if (getenv("LD_PRELOAD"))
    unsetenv("MALLOC_TRACE");
  mcheck(NULL);
# endif
#endif
}

#if TRACE_MEMORY
PUBLIC void MEMORY_alloc(void *p_ptr, size_t size, const char *src)
#else
PUBLIC void MEMORY_alloc(void *p_ptr, size_t size)
#endif
{
  void *alloc;

  alloc = malloc(size);

  if (!alloc)
    THROW(E_MEMORY);

  *((void **)p_ptr) = alloc;
  MEMORY_count++;

  #if TRACE_MEMORY
  TRACE_where();
  printf("%s: MEMORY_alloc(%d) -> %p\n", src, size, alloc);
  fflush(stdout);
  #endif
}


#if TRACE_MEMORY
PUBLIC void MEMORY_alloc_zero(void *p_ptr, size_t size, const char *src)
#else
PUBLIC void MEMORY_alloc_zero(void *p_ptr, size_t size)
#endif
{
  void *alloc;

  alloc = calloc(size, 1);

  if (!alloc)
    THROW(E_MEMORY);

  *((void **)p_ptr) = alloc;
  MEMORY_count++;

  #if TRACE_MEMORY
  TRACE_where();
  printf("%s: MEMORY_alloc_zero(%d) -> %p\n", src, size, alloc);
  fflush(stdout);
  #endif
}


#if TRACE_MEMORY
PUBLIC void MEMORY_realloc(void *p_ptr, size_t size, const char *src)
#else
PUBLIC void MEMORY_realloc(void *p_ptr, size_t size)
#endif
{
  void *alloc = *((void **)p_ptr);

  alloc = realloc(alloc, size);

  if (!alloc)
    THROW(E_MEMORY);

  #if TRACE_MEMORY
  TRACE_where();
  printf("%s: MEMORY_realloc(%p, %d) -> %p\n", src, *((void **)p_ptr), size, alloc);
  fflush(stdout);
  #endif

  *((void **)p_ptr) = alloc;
}


#if TRACE_MEMORY
PUBLIC void MEMORY_free(void *p_ptr, const char *src)
#else
PUBLIC void MEMORY_free(void *p_ptr)
#endif
{
  void *alloc = *((void **)p_ptr);

  *((char *)alloc) = 0x23;

  free(alloc);

  *((void **)p_ptr) = NULL;
  MEMORY_count--;

  #if TRACE_MEMORY
  TRACE_where();
  printf("%s: MEMORY_free(%p)\n", src, alloc);
  fflush(stdout);
  #endif
}

